<?php

namespace core\database;
use core\{utility\Utility};

class Database {
    private $driver;
    private static $instance;

    public $query = '';
    public $table = '';
    public $query_type = '';
    public $data = '';
    public $data_raw = [];
    public $where = [];
    public $sort = [];
    public $limit = '';
    public $last_id = 0;
    public $result;

    public $prefix = '';

	static function get($new = false, $settings = []) {
		if (!self::$instance || $new) {
			$settings = $settings ?: Utility::get_settings()['database'];

			self::$instance = new self;
			self::$instance->prefix = $settings['prefix'] ?? '';
			self::$instance->driver = (new Mysql)->get($settings);
		}

        return self::$instance;
    }

    static function clone($object = null) {
		$object = $object ?: self::$instance;

		$instance = clone $object;
		$instance->driver = new Mysql;

		return $instance;
	}

	// common

    function set_errors() {
        $errors = $this->driver->errors;

        if ($errors) {
			Utility::set_errors('database', $errors, 'merge');
		}
	}

	static function check() {
		$result = EQSASH_INSTALLED && !Utility::get_errors('database');

		return $result;
	}

	function reset() {
		$this->query = $this->table = $this->data = $this->limit = '';
		$this->result = $this->where = $this->sort = $this->data_raw = [];

		$this->last_id = 0;
		$this->query_type = '';
	}

    function escape($value) {
		if (is_string($value)) {
			$escape = $this->driver->escape($value);
			$value = sprintf('%s', $escape);
		}

		return $value;
	}

    static function query_common($type, $table) {
		$instance = self::get(true);
		$instance->reset();

		$table = $instance->prefix ? $instance->prefix.'_'.$table : $table;
		$table = $instance->escape($table);

		$instance->query = $instance->driver::$queries[$type].' `'.$table.'`';
		$instance->table = $table;
		$instance->query_type = $type;

        return $instance;
    }

	static function query_raw($data) {
		$instance = self::get(true);
		$instance->reset();

		$query = trim(array_shift($data));

		$back_position = mb_strpos($query, '`') + 1;
		$table = mb_substr($query, $back_position, mb_strpos($query, '`', $back_position) - $back_position);
		$query = $instance->prefix ? substr_replace($query, $instance->prefix.'_', $back_position, 0) : $query;

		$conditions = [];
		foreach ($data as $key => $value) {
			$conditions[] = $instance->escape($value);
		}

		$instance->query = vsprintf($query, $conditions);
		$instance->table = $table;
		$instance->query_type = mb_strtolower(mb_substr($query, 0, mb_strpos($query, ' ')));

		return $instance;
	}

	// operations

    static function create($table) {
        return self::query_common(__FUNCTION__, $table);
    }

    static function drop($table) {
        return self::query_common(__FUNCTION__, $table);
    }

    static function select($table) {
        return self::query_common(__FUNCTION__, $table);
    }

    static function insert($table) {
        return self::query_common(__FUNCTION__, $table);
    }

    static function update($table) {
        return self::query_common(__FUNCTION__, $table);
    }

    static function delete($table) {
        return self::query_common(__FUNCTION__, $table);
    }

	// conditions

    function data($data) {
		if ($this->query_type == 'create') {
			return $this->data_create($data);
		} else {
			foreach ($data as $key => $value) {
				$key = $this->escape($key);
				$value = $this->escape($value);

				if ($key == 'id') {
					$this->last_id = $value;
				}

				$data[$key] = $value;
			}

			$data_string = $this->driver->data($data, $this->query_type);

			$this->data = $data_string;
			$this->data_raw = $data;

			return $this->get();
		}
    }

    function data_create($data) {
		foreach ($data as $key => $value) {
			$key = $this->escape($key);

			foreach ($value as $key_two => $value_two) {
				$key_two = $this->escape($key_two);
				$value_two = $this->escape($value_two);

				$value[$key_two] = $value_two;
			}

			$data[$key] = $value;
		}

		$this->data = $this->driver->data_create($data);

		return $this->get();
	}

    function where($key, $value, $operator = '=', $glue = 'AND') {
		$key = $this->escape($key);

		if (is_array($value)) {
			foreach ($value as $key_two => $value_two) {
				$value_two = $this->escape($value_two);
			}
		} else {
			$value = $this->escape($value);

			if ($key == 'id') {
				$this->last_id = $value;
			}
		}
		
		$operators_stop = ['AND', 'OR',];
		$operator = in_array($operator, $operators_stop) ? '=' : $operator;

		$this->where = $this->driver->where($key, $value, $operator, $glue, $this->where);

        return $this->get();
    }

    function sort($key, $value = 'ASC') {
		$this->sort = $this->driver->sort($key, $value);

        return $this->get();
    }

    function limit($key, $value = 0) {
		$this->limit = $this->driver->limit($key, $value);

        return $this->get();
    }

	// execute

    function build() {
		$this->query = $this->driver->build($this);

		if ($this->query_type == 'select') {
			$this->query = str_replace('[FIELDS]', $this->driver->get_select_fields(), $this->query);
		}
	}

    function execute() {
		if (!in_array($this->query_type, ['drop'])) {
			$this->build();
		}

		$result = $this->driver->query($this->query);
		$this->result = $result;

		if ($this->query_type == 'delete' && !$this->where) {
			$result = $this->driver->clear_increment($this->table);
		}

		$result = $result ? true : false;

		$this->set_errors();

		return $result;
	}

    function fetch($type = 'object_all', $index = '', $connection = false) {
		$types = [
			'object_all' => [],
			'object' => [],
			'assoc' => [],
			'assoc_all' => [],
			'assoc_map' => [],
			'assoc_map_all' => [],
			'count' => 0,
			'field' => '',
			'connection' => (object)[],
		];

		$result = $types[$type];

		if ($type == 'count') {
			$this->driver->set_select_fields('count');
		}

		$this->execute();

		if ($this->result && $this->query_type == 'select') {
			if ($this->driver->get_rows($this->result) > 0) {
				if ($type == 'object') {
					$result = $this->driver->fetch_object($this->result);
				} else if ($type == 'object_all') {
					$count = 0;
					while ($row = $this->driver->fetch_object($this->result)) {
						$index_value = $row->$index ?? $count;
						$result[$index_value] = $row;

						$count++;
					}
				} else if ($type == 'assoc') {
					$result = $this->driver->fetch_assoc($this->result);
				} else if ($type == 'assoc_all') {
					$rows = $this->driver->fetch_all($this->result);
					$result = $rows;

					if ($index) {
						$result = [];
						foreach ($rows as $key => $value) {
							$index_value = $value[$index] ?? $key;
							$result[$index_value] = $value;
						}
					}
				} else if ($type == 'assoc_map' || $type == 'assoc_map_all') {
					$rows = $this->driver->fetch_all($this->result);
					foreach ($rows as $key => $value) {
						if ($this->data_raw) {
							$values_count = count($this->data_raw);

							if (!$values_count) {
								break;
							} else if ($values_count == 1) {
								$result[] = reset($value);
							} else {
								$a = reset($value);
								$b = next($value);
								$b = $b === false ? $a : $b;

								$result[$a] = $b;
							}
						} else {
							$a = reset($value);
							$b = next($value);

							if ($a !== false) {
								if ($b !== false) {
									$result[$a] = $b;
								} else {
									$result[] = $a;
								}
							}
						}

						if ($type == 'assoc_map') {
							break;
						}
					}
				} else if ($type == 'field' || $type == 'count') {
					$row = $this->driver->fetch_assoc($this->result);
					$result = $row ? reset($row) : ($type == 'field' ? '' : 0);
				}
			}
		} else if ($this->result) {
			$result = $this->driver;
			if (!$connection) {
				if ($this->query_type == 'insert') {
					$result = $this->driver->get_insert_id() ?? 0;
				} else if (in_array($this->query_type, ['update', 'delete'])) {
					$result = $this->last_id;
				}
			}
		}

		if ($type == 'connection') {
			$result = $this->driver;
		}

		return $result;
    }
}
